"""SMA WebConnect library for Python.

See: http://www.sma.de/en/products/monitoring-control/webconnect.html

Source: http://www.github.com/kellerza/pysma
"""
import asyncio
import copy
import json
import logging
from typing import Any, Dict, Optional

import jmespath  # type: ignore
from aiohttp import ClientSession, ClientTimeout, client_exceptions, hdrs

from . import definitions
from .const import (
    DEFAULT_LANG,
    DEFAULT_TIMEOUT,
    DEVICE_INFO,
    ENERGY_METER_VIA_INVERTER,
    FALLBACK_DEVICE_INFO,
    GENERIC_SENSORS,
    OPTIMIZERS_VIA_INVERTER,
    URL_ALL_PARAMS,
    URL_ALL_VALUES,
    URL_DASH_LOGGER,
    URL_DASH_VALUES,
    URL_LOGGER,
    URL_LOGIN,
    URL_LOGOUT,
    URL_VALUES,
    USERS,
)
from .exceptions import (
    SmaAuthenticationException,
    SmaConnectionException,
    SmaReadException,
)
from .helpers import version_int_to_string
from .sensor import Sensors

_LOGGER = logging.getLogger(__name__)


class SMA:
    """Class to connect to the SMA webconnect module and read parameters."""

    # pylint: disable=too-many-instance-attributes
    _aio_session: ClientSession
    _new_session_data: Optional[dict]
    _url: str
    _sid: Optional[str]
    _uid: Optional[str]
    _lang: str
    _l10n: Optional[dict]
    _devclass: Optional[str]
    _device_info_sensors: Sensors

    def __init__(
        self,
        session: ClientSession,
        url: str,
        password: Optional[str] = None,
        group: str = "user",
        uid: Optional[str] = None,
        lang: str = DEFAULT_LANG,
    ):
        """Init SMA connection.

        Args:
            session (ClientSession): aiohttp client session
            url (str): Url or IP address of device
            password (str, optional): Password to use during login. Defaults to None.
            group (str, optional): Username to use during login. Defaults to "user".
            uid (str, optional): uid used for data extraction. Defaults to None.
            lang (str, optional): Language code of file to retrieve. Defaults to "en-US".

        Raises:
            KeyError: User was not in USERS
        """
        # pylint: disable=too-many-arguments
        if group not in USERS:
            raise KeyError(f"Invalid user type: {group}")
        if password is not None and len(password) > 12:
            _LOGGER.warning("Password should not exceed 12 characters")
        if password is None:
            self._new_session_data = None
        else:
            self._new_session_data = {"right": USERS[group], "pass": password}
        self._url = url.rstrip("/")
        if not url.startswith("http"):
            self._url = "http://" + self._url
        self._aio_session = session
        self._sid = None
        self._uid = uid
        self._lang = lang
        self._l10n = None
        self._devclass = None
        self._device_info_sensors = Sensors(definitions.sensor_map[DEVICE_INFO])

    async def _request_json(
        self, method: str, url: str, **kwargs: Dict[str, Any]
    ) -> dict:
        """Request json data for requests.

        Args:
            method (str): HTTP method to use
            url (str): URL to do request to

        Raises:
            SmaConnectionException: Connection to device failed

        Returns:
            dict: json returned by device
        """
        if self._sid:
            kwargs.setdefault("params", {})
            kwargs["params"]["sid"] = self._sid

        _LOGGER.debug("Sending %s request to %s: %s", method, url, kwargs)

        max_retries = 2
        for retry in range(max_retries):
            try:
                async with self._aio_session.request(
                    method,
                    self._url + url,
                    timeout=ClientTimeout(total=DEFAULT_TIMEOUT),
                    **kwargs,
                ) as res:
                    res_json = await res.json()
                    _LOGGER.debug("Received reply %s", res_json)
                    return res_json or {}
            except (client_exceptions.ContentTypeError, json.decoder.JSONDecodeError):
                _LOGGER.warning("Request to %s did not return a valid json.", url)
                break
            except client_exceptions.ServerDisconnectedError as exc:
                if (retry + 1) < max_retries:
                    # For some reason the SMA device sometimes raises a server disconnected error
                    # If this happens we will retry up to `max_retries` times
                    # This prevents errors in Home Assistant
                    _LOGGER.debug("ServerDisconnectedError, will retry connection.")
                    continue

                raise SmaConnectionException(
                    f"Server at {self._url} disconnected {max_retries+1} times."
                ) from exc
            except (
                client_exceptions.ClientError,
                asyncio.exceptions.TimeoutError,
            ) as exc:
                raise SmaConnectionException(
                    f"Could not connect to SMA at {self._url}: {exc}"
                ) from exc

        return {}

    async def _get_json(self, url: str) -> dict:
        """Get json data for requests.

        Args:
            url (str): URL to do GET request to

        Returns:
            dict: json returned by device
        """
        return await self._request_json(hdrs.METH_GET, url)

    async def _post_json(self, url: str, payload: Optional[dict] = None) -> dict:
        """Post json data for requests.

        Args:
            url (str): URL to do POST request to
            payload (dict, optional): payload to send to device. Defaults to empty dict.

        Returns:
            dict: json returned by device
        """
        if payload is None:
            payload = {}

        params: Dict[str, Any] = {
            "data": json.dumps(payload),
            "headers": {"content-type": "application/json"},
        }

        return await self._request_json(hdrs.METH_POST, url, **params)

    async def _read_l10n(self) -> dict:
        """Read device language file. Returns cached value on subsequent calls.

        Returns:
            dict: json returned by device
        """
        if self._l10n is None:
            self._l10n = await self._get_json(f"/data/l10n/{self._lang}.json")
            if len(self._l10n) == 0:
                _LOGGER.warning(
                    "Language '%s' not supported, fallback to '%s'",
                    self._lang,
                    DEFAULT_LANG,
                )
                self._l10n = await self._get_json(f"/data/l10n/{DEFAULT_LANG}.json")
        return self._l10n

    async def _read_body(self, url: str, payload: dict) -> dict:
        """Parse the json returned by the device and extract result.

        Args:
            url (str): URL to reqquest data from
            payload (dict): payload to send to device
        Raises:
            SmaReadException: The json returned by the device was in an unexpected format

        Returns:
            dict: json result
        """
        if self._sid is None and self._new_session_data is not None:
            await self.new_session()
        body = await self._post_json(url, payload)

        # On the first error we close the session which will re-login
        err = body.get("err")

        if err is not None:
            _LOGGER.warning(
                "%s: error detected, closing session to force another login attempt, got: %s",
                self._url,
                body,
            )
            await self.close_session()
            raise SmaReadException("Error detected while reading")

        if "result" not in body:
            _LOGGER.warning("No 'result' in reply from SMA, got: %s", body)
            raise SmaReadException("No 'result' in reply from SMA")

        if self._uid is None:
            # Get the unique ID
            self._uid = next(iter(body["result"].keys()), None)

        result_body = body["result"].pop(self._uid, None)
        if body != {"result": {}}:
            _LOGGER.warning(
                "Unexpected body %s, extracted %s",
                json.dumps(body),
                json.dumps(result_body),
            )

        return result_body

    async def new_session(self) -> bool:
        """Establish a new session.

        Raises:
            SmaAuthenticationException: Authentication failed

        Returns:
            bool: authentication successful
        """
        body = await self._post_json(URL_LOGIN, self._new_session_data)
        self._sid = jmespath.search("result.sid", body)
        if self._sid:
            _LOGGER.debug("New SID: %s", self._sid)
            return True

        err = body.pop("err", None)
        msg = "Could not start session: %s"

        if err:
            if err == 503:
                _LOGGER.error(msg, "Max amount of sessions reached")
            elif err == 404:
                if not self._url.startswith("https"):
                    _LOGGER.error(msg, "Login URL not found, try using HTTPS")
                else:
                    _LOGGER.error(msg, "Login URL not found")
            else:
                _LOGGER.error(msg, err)
        else:
            _LOGGER.error(msg, "Session ID expected [result.sid]")

        raise SmaAuthenticationException()

    async def close_session(self) -> None:
        """Close the session login."""
        if self._sid is None:
            return
        try:
            await self._post_json(URL_LOGOUT)
        finally:
            self._sid = None

    async def read(self, sensors: Sensors) -> bool:
        """Read a set of keys.

        Args:
            sensors (Sensors): Sensors object containing Sensor objects to read

        Returns:
            bool: reading was successful
        """
        if self._new_session_data is None:
            payload: Dict[str, Any] = {"destDev": [], "keys": []}
            result_body = await self._read_body(URL_DASH_VALUES, payload)
        else:
            payload = {
                "destDev": [],
                "keys": list({s.key for s in sensors if s.enabled}),
            }
            result_body = await self._read_body(URL_VALUES, payload)

        notfound = []
        l10n = await self._read_l10n()
        for sen in sensors:
            if sen.enabled:
                if sen.key in result_body:
                    sen.extract_value(result_body, l10n)
                    continue

                notfound.append(f"{sen.name} [{sen.key}]")

        if notfound:
            _LOGGER.info(
                "No values for sensors: %s. Response from inverter: %s",
                ",".join(notfound),
                result_body,
            )

        return True

    async def read_dash_logger(self) -> dict:
        """Read the dash loggers.

        Returns:
            dict: Dictionary containing loggers returned by device.
        """
        return await self._read_body(URL_DASH_LOGGER, {"destDev": [], "key": []})

    async def read_logger(self, log_id: int, start: int, end: int) -> list:
        """Read a logging key and return the results.

        Args:
            log_id (int): The ID of the log to read.
                totWhOut5min: 28672
                totWhOutDaily: 28704
                GridMsTotWhOutDaily: 28752
                GridMsTotWhInDaily: 28768
                ObjLogBatCha: 28816
                totWhIn5min: 28736
                totWhInDaily: 28768
                ObjLogBatChrg: 29344
                ObjLogBatDsch: 29360
            start (int): Start timestamp in seconds.
            end (int): End timestamp in seconds.

        Returns:
            list: The log entries returned by the device
        """
        payload = {
            "destDev": [],
            "key": log_id,
            "tStart": start,
            "tEnd": end,
        }
        result_body = await self._read_body(URL_LOGGER, payload)
        if not isinstance(result_body, list):
            raise SmaReadException("List of log entries expected.")

        return result_body

    async def device_info(self) -> dict:
        """Read device info and return the results.

        Returns:
            dict: dict containing serial, name, type, manufacturer and sw_version
        """
        await self.read(self._device_info_sensors)

        device_info = {
            "serial": self._device_info_sensors["serial_number"].value
            or FALLBACK_DEVICE_INFO["serial"],
            "name": self._device_info_sensors["device_name"].value
            or FALLBACK_DEVICE_INFO["name"],
            "type": self._device_info_sensors["device_type"].value
            or FALLBACK_DEVICE_INFO["type"],
            "manufacturer": self._device_info_sensors["device_manufacturer"].value
            or FALLBACK_DEVICE_INFO["manufacturer"],
            "sw_version": version_int_to_string(
                self._device_info_sensors["device_sw_version"].value
            ),
        }

        return device_info

    async def _read_all_sensors(self) -> dict:
        all_values = await self._read_body(URL_ALL_VALUES, {"destDev": []})
        all_params = await self._read_body(URL_ALL_PARAMS, {"destDev": []})
        return all_values | all_params

    async def get_sensors(self) -> Sensors:
        """Get the sensors that are present on the device.

        Returns:
            Sensors: Sensors object containing Sensor objects
        """
        all_sensors = await self._read_all_sensors()
        sensor_keys = all_sensors.keys()
        device_sensors = Sensors()

        _LOGGER.debug("Matching generic sensors")

        for sensor in definitions.sensor_map[GENERIC_SENSORS]:
            if sensor.key in sensor_keys:
                if isinstance(all_sensors[sensor.key], list):
                    # SB1.5 multi value
                    val_len = len(all_sensors[sensor.key])
                elif "val" in all_sensors[sensor.key]:
                    # SB1.5 single value
                    val_len = 1
                else:
                    # All other types, single or multi value
                    sensors_values = list(all_sensors[sensor.key].values())[0]
                    val_len = len(sensors_values)
                _LOGGER.debug("Found %s with %d value(s).", sensor.key, val_len)

                if sensor.key_idx < val_len:
                    _LOGGER.debug(
                        "Adding sensor %s (%s_%s)",
                        sensor.name,
                        sensor.key,
                        sensor.key_idx,
                    )
                    device_sensors.add(sensor)

        _LOGGER.debug("Checking if Energy Meter is present...")
        # Detect and add Energy Meter sensors
        em_sensor = copy.copy(definitions.energy_meter)
        em_sensor.extract_value(all_sensors)

        if em_sensor.value:
            _LOGGER.debug(
                "Energy Meter with serial %s detected. Adding extra sensors.",
                em_sensor.value,
            )
            device_sensors.add(
                [
                    sensor
                    for sensor in definitions.sensor_map[ENERGY_METER_VIA_INVERTER]
                    if sensor not in device_sensors
                ]
            )

        _LOGGER.debug("Finding connected optimizers...")
        # Detect and add Optimizer Sensors
        optimizers = all_sensors.get(definitions.optimizer_serial.key)
        if optimizers:
            serials = optimizers.popitem()[1]

            for idx, serial in enumerate(serials or []):
                if serial["val"]:
                    _LOGGER.debug(
                        "Optimizer %s with serial %s detected. Adding extra sensors.",
                        idx,
                        serial,
                    )
                    for sensor_definition in definitions.sensor_map[
                        OPTIMIZERS_VIA_INVERTER
                    ]:
                        new_sensor = copy.copy(sensor_definition)
                        new_sensor.key_idx = idx
                        new_sensor.name = f"{sensor_definition.name}_{idx}"
                        device_sensors.add(new_sensor)

        return device_sensors



"""Test pysma helpers file."""
from pysma.helpers import version_int_to_string


def test_version_int_to_string():
    """Ensure version_int_to_string returns correct values."""
    assert version_int_to_string(4294967294) == "ff.ff.255."
    assert version_int_to_string(51387396) == "3.10.28.R"
    assert version_int_to_string(1) == "0.0.0.E"
    assert version_int_to_string(0) == ""
    assert version_int_to_string(None) == ""



"""Sensor classes for SMA WebConnect library for Python."""
import copy
import logging
from typing import Any, Iterator, List, Optional, Union

import attr
import jmespath  # type: ignore

from .const import JMESPATH_VAL, JMESPATH_VAL_IDX, JMESPATH_VAL_STR

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class Sensor:
    """pysma sensor."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods
    key: str = attr.ib()
    name: str = attr.ib()
    unit: str = attr.ib(default=None)
    factor: int = attr.ib(default=None)
    path: Union[list, tuple] = attr.ib(default=None)
    enabled: bool = attr.ib(default=True)
    l10n_translate: bool = attr.ib(default=False)
    value: Any = attr.ib(default=None, init=False)
    key_idx: int = attr.ib(default=0, repr=False, init=False)

    def __attrs_post_init__(self) -> None:
        """Post init Sensor."""
        key = str(self.key)
        skey = key.split("_")
        if len(skey) > 2 and skey[2].isdigit():
            self.key = f"{skey[0]}_{skey[1]}"
            self.key_idx = int(skey[2])

    def extract_value(self, result_body: dict, l10n: Optional[dict] = None) -> bool:
        """Extract value from json body.

        Args:
            result_body (dict): json body retrieved from device
            l10n (dict, optional): Dictionary to translate tags to strings. Defaults to None.

        Returns:
            bool: Extracting value successful
        """
        try:
            res = result_body[self.key]
        except (KeyError, TypeError):
            _LOGGER.warning("Sensor %s: Not found in %s", self.key, result_body)
            res = self.value
            self.value = None
            return self.value != res

        if not isinstance(self.path, str):
            # Try different methods until we can decode...
            _paths = (
                [sens_path.format(self.key_idx) for sens_path in self.path]
                if isinstance(self.path, (list, tuple))
                else [
                    JMESPATH_VAL,
                    JMESPATH_VAL_STR.format(self.key_idx),
                    JMESPATH_VAL_IDX.format(self.key_idx),
                ]
            )

            while _paths:
                _path = _paths.pop()
                _val = jmespath.search(_path, res)
                if _val is not None:
                    _LOGGER.debug(
                        "Sensor %s: Will be decoded with %s from %s",
                        self.name,
                        _path,
                        res,
                    )
                    self.path = _path
                    break

        # Extract new value
        if isinstance(self.path, str):
            res = jmespath.search(self.path, res)
        else:
            _LOGGER.debug(
                "Sensor %s: No successful value decoded yet: %s", self.name, res
            )
            res = None

        # SMA will return None instead of 0 if if no power is generated
        # For "W" sensors we will set it to 0 by default.
        if res is None and self.unit == "W":
            res = 0

        if isinstance(res, (int, float)) and self.factor:
            res /= self.factor

        if self.l10n_translate and isinstance(l10n, dict):
            res = l10n.get(
                str(res),
                res,
            )

        try:
            return res != self.value
        finally:
            self.value = res


class Sensors:
    """SMA Sensors."""

    def __init__(self, sensors: Union[Sensor, List[Sensor], None] = None):
        """Init Sensors.

        Args:
            sensors (Union[Sensor, List[Sensor], None], optional): One or a list of sensors
                to add on init. Defaults to None.
        """
        self.__s: List[Sensor] = []

        if sensors:
            self.add(sensors)

    def __len__(self) -> int:
        """Length of Sensor list.

        Returns:
            int: Number of Sensor objects
        """
        return len(self.__s)

    def __contains__(self, key: Union[str, Sensor]) -> bool:
        """Check if a sensor is defined.

        Args:
            key (str, Sensor): [description]

        Returns:
            bool: [description]
        """
        if isinstance(key, Sensor):
            key = key.name

        try:
            if self[key]:
                return True
        except KeyError:
            pass
        return False

    def __getitem__(self, key: str) -> Sensor:
        """Get a sensor.

        Args:
            key (str): Either the name or key of the Sensor

        Raises:
            KeyError: Item was not found

        Returns:
            Sensor: The matching Sensor object
        """
        for sen in self.__s:
            if key in (sen.name, sen.key):
                return sen
        raise KeyError(key)

    def __iter__(self) -> Iterator[Sensor]:
        """Iterate Sensor objects.

        Yields:
            Iterator[Sensor]: Sensor iterator
        """
        return self.__s.__iter__()

    def add(self, sensor: Union[Sensor, List[Sensor]]) -> None:
        """Add a sensor, logs warning if it exists.

        A copy of sensor, or the sensors in the list, will be added.
        If the sensor name already exists it will be overwritten.
        If the sensor key already exists it will not be added.

        Args:
            sensor (Sensor, List[Sensor]): One or a list of sensors to add

        Raises:
            TypeError: Argument sensor does not match an expected type
        """
        if isinstance(sensor, (list, tuple)):
            for sss in sensor:
                self.add(sss)
            return

        if isinstance(sensor, Sensor):
            sensor = copy.copy(sensor)
        else:
            raise TypeError("pysma.Sensor expected")

        if sensor.name in self:
            old = self[sensor.name]
            self.__s.remove(old)
            _LOGGER.warning("Replacing sensor %s with %s", old, sensor)

        if sensor.key in self and self[sensor.key].key_idx == sensor.key_idx:
            _LOGGER.warning(
                "Duplicate SMA sensor key %s (idx: %s)", sensor.key, sensor.key_idx
            )

        self.__s.append(sensor)



"""Helper functions for the pysma library."""


def version_int_to_string(version_integer: int) -> str:
    """Convert a version integer to a readable string.

    Args:
        version_integer (int): The version integer, as retrieved from the device.

    Returns:
        str: The version translated to a readable string.
    """
    if not version_integer:
        return ""

    appendixes = ["N", "E", "A", "B", "R", "S"]
    version_bytes = version_integer.to_bytes(4, "big")
    version_appendix = (
        appendixes[version_bytes[3]] if 0 <= version_bytes[3] < len(appendixes) else ""
    )
    return f"{version_bytes[0]:x}.{version_bytes[1]:x}.{version_bytes[2]}.{version_appendix}"



"""Test pysma const file."""

import pysma.definitions
from pysma.sensor import Sensor


def test_duplicate_sensors():
    """Test if defined sensors have unique key and name."""
    variables = vars(pysma.definitions)
    found_keys = []
    found_names = []
    for value in variables.values():
        if isinstance(value, Sensor):
            found_key = f"{value.key}_{value.key_idx}"
            found_name = value.name

            assert found_key not in found_keys
            found_keys.append(found_key)

            assert found_name not in found_names
            found_names.append(found_name)


def test_sensor_map():
    """Test if all map entries only contain unique items."""
    for sensor_map in pysma.definitions.sensor_map.values():
        unique_items = list({f"{s.key}_{s.key_idx}": s for s in sensor_map}.values())
        assert unique_items == sensor_map



#!/usr/bin/env python
"""pysma library setup."""
from pathlib import Path

from setuptools import setup

VERSION = "0.7.4"
URL = "https://github.com/kellerza/pysma"

setup(
    name="pysma",
    version=VERSION,
    description="Library to interface an SMA Solar WebConnect module",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url=URL,
    download_url="{}/tarball/{}".format(URL, VERSION),
    author="Johann Kellerman",
    author_email="kellerza@gmail.com",
    license="MIT",
    packages=["pysma"],
    python_requires=">=3.9",
    install_requires=["aiohttp>3.3,<4", "attrs>18", "jmespath<2"],
    zip_safe=True,
)



"""Sensor definitions for SMA WebConnect library for Python."""
from .const import (
    DEVICE_INFO,
    ENERGY_METER_VIA_INVERTER,
    GENERIC_SENSORS,
    JMESPATHS_TAG,
    OPTIMIZERS_VIA_INVERTER,
)
from .sensor import Sensor

# Status - Operation
#: Status of the device
status = Sensor("6180_08214800", "status", path=JMESPATHS_TAG, l10n_translate=True)
#: General operating status
operating_status_general = Sensor(
    "6180_08412800",
    "operating_status_general",
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)

# Status - Operation - Inverter
#: General operating status
inverter_condition = Sensor(
    "6180_08414C00",
    "inverter_condition",
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
#: Inverter Condition
inverter_system_init = Sensor(
    "6800_08811F00",
    "inverter_system_init",
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
#: Grid connection status
grid_connection_status = Sensor(
    "6180_0846A700",
    "grid_connection_status",
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
#: Grid relay status
grid_relay_status = Sensor(
    "6180_08416400",
    "grid_relay_status",
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)

# DC side - DC measurements PV
#: Current power generated by the solar panels (A side)
pv_power_a = Sensor("6380_40251E00_0", "pv_power_a", unit="W")
#: Current power generated by the solar panels (B side)
pv_power_b = Sensor("6380_40251E00_1", "pv_power_b", unit="W")
#: Current power generated by the solar panels (C side)
pv_power_c = Sensor("6380_40251E00_2", "pv_power_c", unit="W", enabled=False)
#: Current voltage generated by the solar panels (A side)
pv_voltage_a = Sensor(
    "6380_40451F00_0", "pv_voltage_a", unit="V", factor=100, enabled=False
)
#: Current voltage generated by the solar panels (B side)
pv_voltage_b = Sensor(
    "6380_40451F00_1", "pv_voltage_b", unit="V", factor=100, enabled=False
)
#: Current voltage generated by the solar panels (C side)
pv_voltage_c = Sensor(
    "6380_40451F00_2", "pv_voltage_c", unit="V", factor=100, enabled=False
)
#: Current amperage generated by the solar panels (A side)
pv_current_a = Sensor("6380_40452100_0", "pv_current_a", unit="A", factor=1000)
#: Current amperage generated by the solar panels (B side)
pv_current_b = Sensor("6380_40452100_1", "pv_current_b", unit="A", factor=1000)
#: Current amperage generated by the solar panels (C side)
pv_current_c = Sensor(
    "6380_40452100_2", "pv_current_c", unit="A", factor=1000, enabled=False
)
#: Isolation Resistance
pv_isolation_resistance = Sensor(
    "6102_00254F00_0",
    "pv_isolation_resistance",
    unit="kOhm",
    factor=1000,
    enabled=False,
)

# DC Side - Insulation monitoring
#: Insulation residual current
insulation_residual_current = Sensor(
    "6102_40254E00", "insulation_residual_current", unit="mA", enabled=False
)

# AC Side - PV Power measurements
#: Current power generated by the solar panels
pv_power = Sensor("6100_0046C200", "pv_power", unit="W", enabled=False)

# AC Side - Grid measurements
#: Power supplied to the grid. grid_power = power_l1 + power_l2 + power_l3
grid_power = Sensor("6100_40263F00", "grid_power", unit="W")
#: Grid frequency
frequency = Sensor("6100_00465700", "frequency", unit="Hz", factor=100, enabled=False)

# AC Side - Grid measurements - Active power
#: Power for phase 1
power_l1 = Sensor("6100_40464000", "power_l1", unit="W", enabled=False)
#: Power for phase 2
power_l2 = Sensor("6100_40464100", "power_l2", unit="W", enabled=False)
#: Power for phase 3
power_l3 = Sensor("6100_40464200", "power_l3", unit="W", enabled=False)

# AC Side - Grid measurements - Reactive power
#: Total Reactive Power
grid_reactive_power = Sensor(
    "6100_40265F00", "grid_reactive_power", unit="var", enabled=False
)
#: Reactive Power for phase 1
grid_reactive_power_l1 = Sensor(
    "6100_40666000", "grid_reactive_power_l1", unit="var", enabled=False
)
#: Reactive Power for phase 2
grid_reactive_power_l2 = Sensor(
    "6100_40666100", "grid_reactive_power_l2", unit="var", enabled=False
)
#: Reactive Power for phase 3
grid_reactive_power_l3 = Sensor(
    "6100_40666200", "grid_reactive_power_l3", unit="var", enabled=False
)

# AC Side - Grid measurements - Apparent power
#: Total Apparent Power
grid_apparent_power = Sensor(
    "6100_40666700", "grid_apparent_power", unit="VA", enabled=False
)
#: Apparent Power for phase 1
grid_apparent_power_l1 = Sensor(
    "6100_40666800", "grid_apparent_power_l1", unit="VA", enabled=False
)
#: Apparent Power for phase 2
grid_apparent_power_l2 = Sensor(
    "6100_40666900", "grid_apparent_power_l2", unit="VA", enabled=False
)
#: Apparent Power for phase 3
grid_apparent_power_l3 = Sensor(
    "6100_40666A00", "grid_apparent_power_l3", unit="VA", enabled=False
)


# AC Side - Grid measurements - Power factor
#: Grid Power factor
grid_power_factor = Sensor(
    "6100_00665900", "grid_power_factor", unit="", factor=1000, enabled=False
)
#: Grid Power factor excitation
grid_power_factor_excitation = Sensor(
    "6180_08465A00",
    "grid_power_factor_excitation",
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)

# AC Side - Grid measurements - Phase Current
#: Current for phase 1
current_l1 = Sensor("6100_40465300", "current_l1", unit="A", factor=1000, enabled=False)
#: Current for phase 2
current_l2 = Sensor("6100_40465400", "current_l2", unit="A", factor=1000, enabled=False)
#: Current for phase 3
current_l3 = Sensor("6100_40465500", "current_l3", unit="A", factor=1000, enabled=False)
#: Total Current
current_total = Sensor("6100_00664F00", "current_total", unit="A", factor=1000)

# AC Side - Grid measurements - Phase voltage
#: Voltage for phase 1
voltage_l1 = Sensor("6100_00464800", "voltage_l1", unit="V", factor=100, enabled=False)
#: Voltage for phase 2
voltage_l2 = Sensor("6100_00464900", "voltage_l2", unit="V", factor=100, enabled=False)
#: Voltage for phase 3
voltage_l3 = Sensor("6100_00464A00", "voltage_l3", unit="V", factor=100, enabled=False)

# AC Side - Measured values - energy
#: Total power yield from a solar installation
total_yield = Sensor("6400_00260100", "total_yield", unit="kWh", factor=1000)
#: The solar plant's yield for today
daily_yield = Sensor("6400_00262200", "daily_yield", unit="Wh")

# AC Side - Measured values - Grid measurements
#: Power supplied to grid measured by energy meter
metering_power_supplied = Sensor("6100_40463600", "metering_power_supplied", unit="W")
#: Power absorbed fromgrid measured by energy meter
metering_power_absorbed = Sensor("6100_40463700", "metering_power_absorbed", unit="W")
#: Grid frequency measured by energy meter
metering_frequency = Sensor(
    "6100_00468100", "metering_frequency", unit="Hz", factor=100
)
#: Total power supplied to the grid measured by energy meter
metering_total_yield = Sensor(
    "6400_00462400", "metering_total_yield", unit="kWh", factor=1000
)
#: Total power from the grid measured by energy meter
metering_total_absorbed = Sensor(
    "6400_00462500", "metering_total_absorbed", unit="kWh", factor=1000
)

# AC Side - Measured values - Phase currents
#: Current for phase 1 measured by energy meter
metering_current_l1 = Sensor(
    "6100_40466500", "metering_current_l1", unit="A", factor=1000
)
#: Current for phase 2 measured by energy meter
metering_current_l2 = Sensor(
    "6100_40466600", "metering_current_l2", unit="A", factor=1000
)
#: Current for phase 3 measured by energy meter
metering_current_l3 = Sensor(
    "6100_40466B00", "metering_current_l3", unit="A", factor=1000
)

# AC Side - Measured values - Phase voltage
#: Voltage for phase 1 measured by energy meter
metering_voltage_l1 = Sensor(
    "6100_0046E500", "metering_voltage_l1", unit="V", factor=100, enabled=False
)
#: Voltage for phase 2 measured by energy meter
metering_voltage_l2 = Sensor(
    "6100_0046E600", "metering_voltage_l2", unit="V", factor=100, enabled=False
)
#: Voltage for phase 3 measured by energy meter
metering_voltage_l3 = Sensor(
    "6100_0046E700", "metering_voltage_l3", unit="V", factor=100, enabled=False
)

# AC Side - Electricity meter - Measured values - Active power feed-in
#: Active Power feed-in for phase 1 measured by energy meter
metering_active_power_feed_l1 = Sensor(
    "6100_0046E800", "metering_active_power_feed_l1", unit="W"
)
#: Active Power feed-in  for phase 2 measured by energy meter
metering_active_power_feed_l2 = Sensor(
    "6100_0046E900", "metering_active_power_feed_l2", unit="W"
)
#: Active Power feed-in  for phase 3 measured by energy meter
metering_active_power_feed_l3 = Sensor(
    "6100_0046EA00", "metering_active_power_feed_l3", unit="W"
)

# AC Side - Electricity meter - Measured values - Active power drawn
#: Active Power drawn for phase 1 measured by energy meter
metering_active_power_draw_l1 = Sensor(
    "6100_0046EB00", "metering_active_power_draw_l1", unit="W"
)
#: Active Power drawn for phase 2 measured by energy meter
metering_active_power_draw_l2 = Sensor(
    "6100_0046EC00", "metering_active_power_draw_l2", unit="W"
)
#: Active Power drawn for phase 3 measured by energy meter
metering_active_power_draw_l3 = Sensor(
    "6100_0046ED00", "metering_active_power_draw_l3", unit="W"
)
#: Current power consumption measured by energy meter
metering_current_consumption = Sensor(
    "6100_00543100", "metering_current_consumption", unit="W", enabled=False
)
#: Total power consumption measured by energy meter
metering_total_consumption = Sensor(
    "6400_00543A00",
    "metering_total_consumption",
    unit="kWh",
    factor=1000,
    enabled=False,
)

# AC Side - PV generation
#: Total kWh generated to date
pv_gen_meter = Sensor("6400_0046C300", "pv_gen_meter", unit="kWh", factor=1000)

# AC Side - Emergency power measurements (Secure Power Supply)
#: Secure power supply voltage
sps_voltage = Sensor(
    "6100_0046C600", "sps_voltage", unit="V", factor=100, enabled=False
)
#: Secure power supply current
sps_current = Sensor(
    "6100_0046C700", "sps_current", unit="A", factor=1000, enabled=False
)
#: Secure power supply power
sps_power = Sensor("6100_0046C800", "sps_power", unit="W", enabled=False)

# PV Inverter Optimizers
#: Serial number of optimizer
optimizer_serial = Sensor("6800_10852600", "optimizer_serial")
#: Power supplied by optimizer
optimizer_power = Sensor("6100_40652A00", "optimizer_power", unit="W")
#: Current supplied by optimizer
optimizer_current = Sensor(
    "6100_40652900", "optimizer_current", unit="A", factor=1000, enabled=False
)
#: Voltage supplied by optimizer
optimizer_voltage = Sensor(
    "6100_40652800", "optimizer_voltage", unit="V", factor=100, enabled=False
)
#: Temperature of optimizer
optimizer_temp = Sensor(
    "6100_40652B00", "optimizer_temp", unit="°C", factor=10, enabled=False
)


# Battery (inverter) - Battery (general parameters)
#: Total battery state of charge
battery_soc_total = Sensor("6100_00295A00", "battery_soc_total", unit="%")
#: State of charge battery A
battery_soc_a = Sensor("6100_00498F00_0", "battery_soc_a", unit="%", enabled=False)
#: State of charge battery B
battery_soc_b = Sensor("6100_00498F00_1", "battery_soc_b", unit="%", enabled=False)
#: State of charge battery C
battery_soc_c = Sensor("6100_00498F00_2", "battery_soc_c", unit="%", enabled=False)
#: Voltage battery A
battery_voltage_a = Sensor(
    "6100_00495C00_0", "battery_voltage_a", unit="V", factor=100, enabled=False
)
#: Voltage battery B
battery_voltage_b = Sensor(
    "6100_00495C00_1", "battery_voltage_b", unit="V", factor=100, enabled=False
)
#: Voltage battery C
battery_voltage_c = Sensor(
    "6100_00495C00_2", "battery_voltage_c", unit="V", factor=100, enabled=False
)
#: Current battery A
battery_current_a = Sensor(
    "6100_40495D00_0", "battery_current_a", unit="A", factor=1000
)
#: Current battery B
battery_current_b = Sensor(
    "6100_40495D00_1", "battery_current_b", unit="A", factor=1000
)
#: Current battery C
battery_current_c = Sensor(
    "6100_40495D00_2", "battery_current_c", unit="A", factor=1000
)
#: Temperature battery A
battery_temp_a = Sensor("6100_40495B00_0", "battery_temp_a", unit="°C", factor=10)
#: Temperature battery B
battery_temp_b = Sensor("6100_40495B00_1", "battery_temp_b", unit="°C", factor=10)
#: Temperature battery C
battery_temp_c = Sensor("6100_40495B00_2", "battery_temp_c", unit="°C", factor=10)
#: Battery status operating mode
battery_status_operating_mode = Sensor(
    "6180_08495E00",
    "battery_status_operating_mode",
    path=JMESPATHS_TAG,
    l10n_translate=True,
)

# Battery (inverter) - Diagnosis
#: Total battery capacity
battery_capacity_total = Sensor("6100_00696E00", "battery_capacity_total", unit="%")
#: Capacity battery A
battery_capacity_a = Sensor(
    "6100_00499100_0", "battery_capacity_a", unit="%", enabled=False
)
#: Capacity battery B
battery_capacity_b = Sensor(
    "6100_00499100_1", "battery_capacity_b", unit="%", enabled=False
)
#: Capacity battery C
battery_capacity_c = Sensor(
    "6100_00499100_2", "battery_capacity_c", unit="%", enabled=False
)

# Battery (inverter) - Charge (voltage)
#: Charging voltage battery A
battery_charging_voltage_a = Sensor(
    "6102_00493500_0", "battery_charging_voltage_a", unit="V", factor=100, enabled=False
)
#: Charging voltage battery B
battery_charging_voltage_b = Sensor(
    "6102_00493500_1", "battery_charging_voltage_b", unit="V", factor=100, enabled=False
)
#: Charging voltage battery C
battery_charging_voltage_c = Sensor(
    "6102_00493500_2", "battery_charging_voltage_c", unit="V", factor=100, enabled=False
)

# Battery (inverter) - Battery charge (power & energy)
#: Total charging power
battery_power_charge_total = Sensor(
    "6100_00496900", "battery_power_charge_total", unit="W"
)
#: Charging power battery A
battery_power_charge_a = Sensor(
    "6100_00499300_0", "battery_power_charge_a", unit="W", enabled=False
)
#: Charging power battery B
battery_power_charge_b = Sensor(
    "6100_00499300_1", "battery_power_charge_b", unit="W", enabled=False
)
#: Charging power battery C
battery_power_charge_c = Sensor(
    "6100_00499300_2", "battery_power_charge_c", unit="W", enabled=False
)
#: Total charge
battery_charge_total = Sensor(
    "6400_00496700", "battery_charge_total", unit="kWh", factor=1000
)
#: Charge battery A
battery_charge_a = Sensor(
    "6400_00499500_0", "battery_charge_a", unit="kWh", factor=1000, enabled=False
)
#: Charge battery B
battery_charge_b = Sensor(
    "6400_00499500_1", "battery_charge_b", unit="kWh", factor=1000, enabled=False
)
#: Charge battery C
battery_charge_c = Sensor(
    "6400_00499500_2", "battery_charge_c", unit="kWh", factor=1000, enabled=False
)

# Battery (inverter) - Battery discharge (power & energy)
#: Total discharging power
battery_power_discharge_total = Sensor(
    "6100_00496A00", "battery_power_discharge_total", unit="W"
)
#: Disharging power battery A
battery_power_discharge_a = Sensor(
    "6100_00499400_0", "battery_power_discharge_a", unit="W", enabled=False
)
#: Disharging power battery B
battery_power_discharge_b = Sensor(
    "6100_00499400_1", "battery_power_discharge_b", unit="W", enabled=False
)
#: Disharging power battery C
battery_power_discharge_c = Sensor(
    "6100_00499400_2", "battery_power_discharge_c", unit="W", enabled=False
)
#: Total discharge
battery_discharge_total = Sensor(
    "6400_00496800", "battery_discharge_total", unit="kWh", factor=1000
)
#: Discharge battery A
battery_discharge_a = Sensor(
    "6400_00499600_0", "battery_discharge_a", unit="kWh", factor=1000, enabled=False
)
#: Discharge battery B
battery_discharge_b = Sensor(
    "6400_00499600_1", "battery_discharge_b", unit="kWh", factor=1000, enabled=False
)
#: Discharge battery C
battery_discharge_c = Sensor(
    "6400_00499600_2", "battery_discharge_c", unit="kWh", factor=1000, enabled=False
)

# Device Parameters
# Type Label - Type Label
#: Device serial number
serial_number = Sensor("6800_00A21E00", "serial_number")
#: Device name
device_name = Sensor("6800_10821E00", "device_name")
#: Device type
device_type = Sensor(
    "6800_08822000", "device_type", path=JMESPATHS_TAG, l10n_translate=True
)
#: Device manufactorer
device_manufacturer = Sensor(
    "6800_08822B00", "device_manufacturer", path=JMESPATHS_TAG, l10n_translate=True
)
#: Device software version
device_sw_version = Sensor("6800_00823400", "device_sw_version")

# Device - Inverter
#: Power limit of the Inverter
inverter_power_limit = Sensor("6800_00832A00", "inverter_power_limit", unit="W")

# System communication - Meter on Speedwire
#: Serial number of energy meter
energy_meter = Sensor("6800_008AA300", "energy_meter")


sensor_map = {
    GENERIC_SENSORS: [
        battery_capacity_a,
        battery_capacity_b,
        battery_capacity_c,
        battery_capacity_total,
        battery_charge_a,
        battery_charge_b,
        battery_charge_c,
        battery_charge_total,
        battery_charging_voltage_a,
        battery_charging_voltage_b,
        battery_charging_voltage_c,
        battery_current_a,
        battery_current_b,
        battery_current_c,
        battery_discharge_a,
        battery_discharge_b,
        battery_discharge_c,
        battery_discharge_total,
        battery_power_charge_a,
        battery_power_charge_b,
        battery_power_charge_c,
        battery_power_charge_total,
        battery_power_discharge_a,
        battery_power_discharge_b,
        battery_power_discharge_c,
        battery_power_discharge_total,
        battery_soc_a,
        battery_soc_b,
        battery_soc_c,
        battery_soc_total,
        battery_status_operating_mode,
        battery_temp_a,
        battery_temp_b,
        battery_temp_c,
        battery_voltage_a,
        battery_voltage_b,
        battery_voltage_c,
        current_l1,
        current_l2,
        current_l3,
        current_total,
        daily_yield,
        frequency,
        grid_apparent_power,
        grid_apparent_power_l1,
        grid_apparent_power_l2,
        grid_apparent_power_l3,
        grid_connection_status,
        grid_power,
        grid_power_factor,
        grid_power_factor_excitation,
        grid_reactive_power,
        grid_reactive_power_l1,
        grid_reactive_power_l2,
        grid_reactive_power_l3,
        grid_relay_status,
        insulation_residual_current,
        inverter_condition,
        inverter_power_limit,
        inverter_system_init,
        operating_status_general,
        power_l1,
        power_l2,
        power_l3,
        pv_current_a,
        pv_current_b,
        pv_current_c,
        pv_gen_meter,
        pv_isolation_resistance,
        pv_power,
        pv_power_a,
        pv_power_b,
        pv_power_c,
        pv_voltage_a,
        pv_voltage_b,
        pv_voltage_c,
        sps_current,
        sps_power,
        sps_voltage,
        status,
        total_yield,
        voltage_l1,
        voltage_l2,
        voltage_l3,
    ],
    OPTIMIZERS_VIA_INVERTER: [
        optimizer_power,
        optimizer_current,
        optimizer_voltage,
        optimizer_temp,
    ],
    ENERGY_METER_VIA_INVERTER: [
        metering_power_supplied,
        metering_power_absorbed,
        metering_frequency,
        metering_total_yield,
        metering_total_absorbed,
        metering_current_l1,
        metering_current_l2,
        metering_current_l3,
        metering_voltage_l1,
        metering_voltage_l2,
        metering_voltage_l3,
        metering_active_power_feed_l1,
        metering_active_power_feed_l2,
        metering_active_power_feed_l3,
        metering_active_power_draw_l1,
        metering_active_power_draw_l2,
        metering_active_power_draw_l3,
        metering_current_consumption,
        metering_total_consumption,
    ],
    DEVICE_INFO: [
        serial_number,
        device_name,
        device_type,
        device_manufacturer,
        device_sw_version,
    ],
}
